/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $USAGI: spd.h,v 1.9 2001/09/12 15:53:50 miyazawa Exp $
 */

#ifndef __SPD_H
#define __SPD_H

#include <linux/list.h>
#include <linux/socket.h>
#include <linux/spinlock.h>
#include <net/sadb.h>

extern struct list_head spd_list;
extern rwlock_t spd_lock;

/* selector holds selector for SPD 
   Selectors are now sorce address and distination address
   Please add selectors in this structure if you want */

struct selector{

	/* TODO: It should be better to hold elements as reference(pointer) */
	struct sockaddr_storage src;
	__u8 prefixlen_s;
	struct sockaddr_storage dst;
	__u8 prefixlen_d;
	__u8   proto;
};

/* ipsec_sp holds selector and sa information.
   sa_entries is a list of SA concerned with this policy. */

#define IPSEC_APPLY   		1
#define IPSEC_DISCARD 		2
#define IPSEC_BYPASS  		3

#define SPD_RLOCK(lock)		read_lock(lock)
#define SPD_RUNLOCK(lock)	read_unlock(lock)
#define SPD_WLOCK(lock)		write_lock(lock)
#define SPD_WUNLOCK(lock)	write_unlock(lock)

/* these are used as lock_type in spd_find_by_selector */
#define SPD_WITH_NOLOCK		0
#define SPD_WITH_RLOCK		1
#define SPD_WITH_WLOCK		2

struct ipsec_sp{
	/* Never touch entry without locking spd_lock */
	struct list_head entry;

	/* This lock only affects elements except for entry. */
	rwlock_t lock;	
	atomic_t count;
	__u8   action;			/* IPSEC_{APPLY, DISCARD, BYPASS} */ 
	struct selector selector;
	struct list_head sa_index_list;
};

struct ipsec_sp* ipsec_sp_kmalloc(void);
int ipsec_sp_init(struct ipsec_sp *policy);
int ipsec_sp_kfree(struct ipsec_sp **policy);
int ipsec_sp_copy(struct ipsec_sp *dst, struct ipsec_sp *src);
int ipsec_sp_release_ref(struct ipsec_sp *policy);
#define spd_release_ref(policy) ipsec_sp_release_ref(policy)
int spd_append(struct ipsec_sp *entry);
int spd_remove(struct selector *selector);
//int spd_update(struct ipsec_sp *policy);
int spd_find_by_selector(struct selector *selector, struct ipsec_sp **policy, int lock_type);
int spd_find_same_spd(struct ipsec_sp *policy, int lock_type);
//int spd_check_sa_list(struct ipsec_sp *policy);
int spd_release_invalid_sa(struct ipsec_sp *policy);
void spd_clear_db(void);

extern int spd_init(void);
extern int spd_cleanup(void);

#endif /* __SPD_H */
