/* $USAGI: pfkey_v2_msg_get.c,v 1.10 2001/09/08 08:32:12 yoshfuji Exp $ */

/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * This is a parse routine for a message of SADB_GET.
 */
#ifdef PFKEY_DEBUG
#define DEBUG 1
#endif

#ifdef MODULE
#include <linux/module.h>
#  ifdef MODVERSIONS
#  include <linux/modversions.h>
#  endif
#endif

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/random.h>

#include <net/pfkeyv2.h>
#include <net/pfkey.h>
#include <net/sadb.h>

#include "pfkey_v2_msg.h"
#include "sadb_utils.h" /* esptoa authtoa */
			/* sockaddrtoa ultoa */
#define BUFSIZE 64

int sadb_msg_get_parse(struct sock *sk, struct sadb_msg *msg, struct sadb_msg **reply)
{
	int error = 0;
	struct sadb_ext *ext_msgs[SADB_EXT_MAX+1], *reply_ext_msgs[SADB_EXT_MAX+1];
        struct sadb_sa *sa = NULL;
        struct sadb_address *src = NULL;
        struct sadb_address *dst = NULL;
        struct ipsec_sa *sa_entry = NULL;
	struct sockaddr_storage saddr, daddr;

	if (!msg) {
		pr_debug("sadb_msg_get_parse: msg==null\n");
		error = -EINVAL;
		goto err;
	}

	memset(reply_ext_msgs, 0, sizeof(reply_ext_msgs));

	memset(ext_msgs, 0, sizeof(ext_msgs));
	error = sadb_msg_detect_ext(msg, ext_msgs);

	if (error) {
		pr_debug("sadb_msg_get_parse: sadb_msg_detect_ext faild\n");
		goto err;
	}

	if (ext_msgs[SADB_EXT_SA] &&
	    ext_msgs[SADB_EXT_ADDRESS_SRC] &&
	    ext_msgs[SADB_EXT_ADDRESS_DST])
	{
		sa  = (struct sadb_sa*)ext_msgs[SADB_EXT_SA];

                src = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_SRC];
                dst = (struct sadb_address*)ext_msgs[SADB_EXT_ADDRESS_DST];

                error = sadb_address_to_sockaddr( src, (struct sockaddr*)&saddr);
                if (error) {
                        pr_debug("sadb_msg_add_parse: src translation failed\n");
                        goto err;
                }
                error = sadb_address_to_sockaddr( dst, (struct sockaddr*)&daddr);
                if (error) {
                        pr_debug("sadb_msg_add_parse: dst translation failed\n");
                        goto err;
                }

		error = sadb_find_by_address_proto_spi((struct sockaddr*)&saddr, src->sadb_address_prefixlen,
					 (struct sockaddr*)&daddr, dst->sadb_address_prefixlen,
					 NULL, 0, /* proxy */
					 msg->sadb_msg_type,
					 sa->sadb_sa_spi,
					 &sa_entry); 


		if (error == -EEXIST) {

			error=pfkey_sa_build(&reply_ext_msgs[SADB_EXT_SA], SADB_EXT_SA,
						sa_entry->spi, sa_entry->replay_window.size,
						sa_entry->state, sa_entry->auth_algo.algo,
						sa_entry->esp_algo.algo, SADB_SAFLAGS_PFS );
			if (error) {
				pr_debug("sadb_msg_get_parse: pfkey_sa_build faild\n");
				goto err;
			}

			error = lifetime_to_sadb_lifetime(&(sa_entry->lifetime_c),
							(struct sadb_lifetime*)reply_ext_msgs[SADB_EXT_LIFETIME_CURRENT],
							SADB_EXT_LIFETIME_CURRENT);
			if (error) {
				pr_debug("sadb_msg_get_parse: lifetime_to_sadb_lifetime failed\n");
				goto err;
			}

			error = lifetime_to_sadb_lifetime(&(sa_entry->lifetime_h),
							(struct sadb_lifetime*)reply_ext_msgs[SADB_EXT_LIFETIME_HARD],
							SADB_EXT_LIFETIME_HARD);
			if (error) {
				pr_debug("sadb_msg_get_parse: lifetime_to_sadb_lifetime failed\n");
				goto err;
			}

			error = lifetime_to_sadb_lifetime(&(sa_entry->lifetime_s),
							(struct sadb_lifetime*)reply_ext_msgs[SADB_EXT_LIFETIME_SOFT],
							SADB_EXT_LIFETIME_SOFT);
			if (error) {
				pr_debug("sadb_msg_get_parse: lifetime_to_sadb_lifetime failed\n");
				goto err;
			}

			error = pfkey_address_build((struct sadb_ext**)&reply_ext_msgs[SADB_EXT_ADDRESS_SRC],
						SADB_EXT_ADDRESS_SRC,
						sa_entry->proto,
						sa_entry->prefixlen_s,
						(struct sockaddr*)&(sa_entry->src));
			if (error) {
				pr_debug("sadb_msg_get_parse: pfkey_address_build faild\n");
				goto err;
			}

			error = pfkey_address_build((struct sadb_ext**)&reply_ext_msgs[SADB_EXT_ADDRESS_DST],
						SADB_EXT_ADDRESS_DST,
						sa_entry->proto,
						sa_entry->prefixlen_d,
						(struct sockaddr*)&(sa_entry->dst));
			if (error) {
				pr_debug("sadb_msg_get_parse: pfkey_address_build faild\n");
				goto err;
			}

			error = pfkey_address_build((struct sadb_ext**)&reply_ext_msgs[SADB_EXT_ADDRESS_PROXY],
						SADB_EXT_ADDRESS_PROXY,
						sa_entry->proto,
						sa_entry->prefixlen_p,
						(struct sockaddr*)&(sa_entry->pxy));
			if (error) {
				pr_debug("sadb_msg_get_parse: pfkey_address_build faild\n");
				goto err;
			}

			if (sa_entry->ipsec_proto == SADB_SATYPE_AH) {
				error = pfkey_key_build((struct sadb_ext**)&reply_ext_msgs[SADB_EXT_KEY_AUTH],
						SADB_EXT_KEY_AUTH,
						sa_entry->auth_algo.key_len*sizeof(__u8),
						sa_entry->auth_algo.key);
				if (error) {
					pr_debug("sadb_msg_get_parse: pfkey_key_build faild\n");
					goto err;
				}
			}

			if (sa_entry->ipsec_proto == SADB_SATYPE_ESP) {
				error = pfkey_key_build((struct sadb_ext**)&reply_ext_msgs[SADB_EXT_KEY_ENCRYPT],
						SADB_EXT_KEY_ENCRYPT,
						sa_entry->esp_algo.key_len*sizeof(__u8),
						sa_entry->esp_algo.key);
				if (error) {
					pr_debug("sadb_msg_get_parse: pfkey_key_build faild\n");
					goto err;
				}
			}
			
		} else {
			pr_debug("sadb_msg_get_parse: sadb_find_by_address_spi faild\n");
			goto err;
		}
	}

	sadb_release_ref(sa_entry);
//	sadb_msg_reply_build(msg, reply_ext_msgs, reply);
	pfkey_extensions_free(reply_ext_msgs);
	return error;
err:
//	sadb_msg_reply_build2(msg, error, reply);
	return error;
}

